/*
 *
 *   Copyright (C) 2005-2010 by Raymond Huang
 *   plushuang at users.sourceforge.net
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *  ---
 *
 *  In addition, as a special exception, the copyright holders give
 *  permission to link the code of portions of this program with the
 *  OpenSSL library under certain conditions as described in each
 *  individual source file, and distribute linked combinations
 *  including the two.
 *  You must obey the GNU Lesser General Public License in all respects
 *  for all of the code used other than OpenSSL.  If you modify
 *  file(s) with this exception, you may extend this exception to your
 *  version of the file(s), but you are not obligated to do so.  If you
 *  do not wish to do so, delete this exception statement from your
 *  version.  If you delete this exception statement from all source
 *  files in the program, then also delete it here.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <uglib.h>
#include <ug_data_download.h>
#include <ug_category-gtk.h>
#include <glib/gi18n.h>

// static functions
static void		ug_category_gtk_init     (UgCategoryGtk* cgtk);
static void		ug_category_gtk_finalize (UgCategoryGtk* cgtk);
static gboolean	ug_category_gtk_watch (UgCategoryGtk* ccmd, UgDataset* dataset, UgMessage* message, gpointer user_data);
// Function used by GtkTreeModelSort.
static gint		ug_download_model_compare (GtkTreeModel* model, GtkTreeIter* a, GtkTreeIter* b, gpointer user_data);
// Function used by GtkTreeModelFilter.
static gboolean	ug_download_model_filter_queuing  (GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static gboolean	ug_download_model_filter_finished (GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static gboolean	ug_download_model_filter_recycled (GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
// Function used by ug_category_gtk_switch().
static UgDataset*	ug_download_model_get_last (GtkTreeModel* model);
static UgDataset*	ug_download_model_get_first (GtkTreeModel* model);

// define UgDataEntry for UgDataClass
static UgDataEntry	category_gtk_data_entry[] =
{
	UG_CATEGORY_DATA_ENTRY,
	{NULL},			// null-terminated
};

static UgDataClass	category_gtk_data_class =
{
	UG_CATEGORY_CLASS_NAME,		// name
	NULL,						// reserve
	sizeof (UgCategoryGtk),		// instance_size
	category_gtk_data_entry,	// entry

	(UgInitFunc)     ug_category_gtk_init,
	(UgFinalizeFunc) ug_category_gtk_finalize,
	(UgAssignFunc)   ug_category_assign,
};
// extern
const	UgDataClass*	UgCategoryGtkClass = &category_gtk_data_class;

static void ug_category_gtk_init (UgCategoryGtk* cgtk)
{
	GtkTreeModel*		model;
	GtkTreeView*		view;
	GtkTreeViewColumn*	column;

	// UgCategory
	ug_category_init ((UgCategory*) cgtk);
	cgtk->watch_func = (UgWatchFunc) ug_category_gtk_watch;
	// UgCategoryGtk
	g_queue_init (&cgtk->active);
	cgtk->store = gtk_list_store_new (1, G_TYPE_POINTER);
	model = GTK_TREE_MODEL (cgtk->store);
	ug_download_widget_init (&cgtk->all,      NULL);
	ug_download_widget_init (&cgtk->queuing,  NULL);
	ug_download_widget_init (&cgtk->finished, NULL);
	ug_download_widget_init (&cgtk->recycled, NULL);

	ug_download_widget_set_sortable (&cgtk->all,    model, ug_download_model_compare);
	ug_download_widget_set_filter (&cgtk->queuing,  model, ug_download_model_filter_queuing);
	ug_download_widget_set_filter (&cgtk->finished, model, ug_download_model_filter_finished);
	ug_download_widget_set_filter (&cgtk->recycled, model, ug_download_model_filter_recycled);
	// All
	view = cgtk->all.view;
	ug_download_view_use_all_icon (view, TRUE);
	// Finished column
	view = cgtk->finished.view;
	column = gtk_tree_view_get_column (view, UG_DOWNLOAD_COLUMN_COMPLETE);
	gtk_tree_view_column_set_visible (column, FALSE);
	column = gtk_tree_view_get_column (view, UG_DOWNLOAD_COLUMN_PERCENT);
	gtk_tree_view_column_set_visible (column, FALSE);
	column = gtk_tree_view_get_column (view, UG_DOWNLOAD_COLUMN_LEFT);
	gtk_tree_view_column_set_visible (column, FALSE);
	column = gtk_tree_view_get_column (view, UG_DOWNLOAD_COLUMN_SPEED);
	gtk_tree_view_column_set_visible (column, FALSE);
	// Recycled column
	view = cgtk->recycled.view;
	column = gtk_tree_view_get_column (view, UG_DOWNLOAD_COLUMN_ELAPSED);
	gtk_tree_view_column_set_visible (column, FALSE);
	column = gtk_tree_view_get_column (view, UG_DOWNLOAD_COLUMN_LEFT);
	gtk_tree_view_column_set_visible (column, FALSE);
	column = gtk_tree_view_get_column (view, UG_DOWNLOAD_COLUMN_SPEED);
	gtk_tree_view_column_set_visible (column, FALSE);
}

static void ug_category_gtk_finalize (UgCategoryGtk* cgtk)
{
	UgDataset*		dataset;
	GtkTreeModel*	model;
	GtkTreeIter		iter;
	gboolean		valid;

	// UgCategory
	ug_category_finalize ((UgCategory*) cgtk);
	// free UgDownloadWidget
	ug_download_widget_finalize (&cgtk->all);
	ug_download_widget_finalize (&cgtk->queuing);
	ug_download_widget_finalize (&cgtk->finished);
	ug_download_widget_finalize (&cgtk->recycled);
	// free all jobs
	model = GTK_TREE_MODEL (cgtk->store);
	valid = gtk_tree_model_get_iter_first (model, &iter);
	while (valid) {
		gtk_tree_model_get (model, &iter, 0, &dataset, -1);
		valid = gtk_tree_model_iter_next (model, &iter);
		ug_category_gtk_remove (cgtk, dataset);
	}
	// free others
	g_queue_clear (&cgtk->active);
	g_object_unref (cgtk->store);
}

UgCategoryGtk*	ug_category_gtk_new  (UgCategoryGtk* primary)
{
	UgCategoryGtk*	cgtk;

	cgtk = ug_data_new (&category_gtk_data_class);
	cgtk->primary = primary;
	return cgtk;
}

void	ug_category_gtk_free (UgCategoryGtk* cgtk)
{
	ug_data_free (cgtk);
}

void	ug_category_gtk_add (UgCategoryGtk* secondary, UgDataset* dataset)
{
	UgCategoryGtk*		primary;
	UgRelationGtk*		relation;

	// add and set UgRelationGtkClass to dataset
	UG_DATASET_RELATION_CLASS (dataset) = UgRelationGtkClass;
	relation = UG_DATASET_RELATION_GTK (dataset);
	if (relation == NULL)
		relation = ug_dataset_alloc_front (dataset, UgRelationGtkClass);
	else {
	// set hint
		if (relation->hints & UG_HINT_ACTIVE)
			relation->hints &= ~UG_HINT_ACTIVE;
	}

	if (secondary->primary)
		primary = secondary->primary;
	else {
		primary = secondary;
		secondary = NULL;
	}
	// add job to primary & secondary category
	if (primary && relation->primary == NULL) {
		relation->primary = primary;
		gtk_list_store_append (primary->store, &relation->primary_iter);
		gtk_list_store_set (primary->store, &relation->primary_iter, 0, dataset, -1);
		ug_dataset_ref (dataset);
	}
	if (secondary && relation->secondary == NULL) {
		relation->secondary = secondary;
		gtk_list_store_append (secondary->store, &relation->secondary_iter);
		gtk_list_store_set (secondary->store, &relation->secondary_iter, 0, dataset, -1);
		ug_dataset_ref (dataset);
	}
}

void	ug_category_gtk_remove (UgCategoryGtk* cgtk, UgDataset* dataset)
{
	UgRelationGtk*		relation;

	relation = UG_DATASET_RELATION_GTK (dataset);
	if (relation->primary) {
		// remove job from primary
		gtk_list_store_remove (relation->primary->store, &relation->primary_iter);
		relation->primary = NULL;
		ug_dataset_unref (dataset);
	}
	if (relation->secondary) {
		cgtk = relation->secondary;
		// stop active job in secondary category
		ug_category_stop ((UgCategory*) cgtk, dataset);
		g_queue_remove (&cgtk->active, dataset);
		// remove from secondary category
		gtk_list_store_remove (cgtk->store, &relation->secondary_iter);
		relation->secondary = NULL;
		// delete data and files
		ug_download_temp_delete (dataset);
		ug_dataset_unref (dataset);
	}
}

void	ug_category_gtk_switch (UgCategoryGtk* cgtk, UgDataset* dataset, UgCategoryHints hint)
{
	UgDataset*		position;
	UgRelationGtk*	reldata;
	UgRelationGtk*	relpos;
	GtkTreeModel*	model;
	GtkTreePath*	path;

	reldata = UG_DATASET_RELATION_GTK (dataset);
	cgtk = reldata->secondary;
	// select queue
	if (hint & UG_HINT_RECYCLED) {
		position = ug_download_model_get_first (cgtk->recycled.model);
		if (position) {
			relpos = UG_DATASET_RELATION_GTK (position);
			gtk_list_store_move_before (reldata->primary->store,
					&reldata->primary_iter, &relpos->primary_iter);
			gtk_list_store_move_before (reldata->secondary->store,
					&reldata->secondary_iter, &relpos->secondary_iter);
		}
		reldata->hints &= ~UG_HINT_FINISHED;
		reldata->hints |=  UG_HINT_RECYCLED;
	}
	else if (hint & UG_HINT_FINISHED) {
		position = ug_download_model_get_first (cgtk->finished.model);
		if (position) {
			relpos = UG_DATASET_RELATION_GTK (position);
			gtk_list_store_move_before (reldata->primary->store,
					&reldata->primary_iter, &relpos->primary_iter);
			gtk_list_store_move_before (reldata->secondary->store,
					&reldata->secondary_iter, &relpos->secondary_iter);
		}
		reldata->hints &= ~UG_HINT_RECYCLED;
		reldata->hints |=  UG_HINT_FINISHED;
	}
	else {
		position = ug_download_model_get_last (cgtk->queuing.model);
		if (position) {
			relpos = UG_DATASET_RELATION_GTK (position);
			gtk_list_store_move_after (reldata->primary->store,
					&reldata->primary_iter, &relpos->primary_iter);
			gtk_list_store_move_after (reldata->secondary->store,
					&reldata->secondary_iter, &relpos->secondary_iter);
		}
		reldata->hints &= ~(UG_HINT_FINISHED | UG_HINT_RECYCLED);
	}
	// emit
	model = GTK_TREE_MODEL (reldata->primary->store);
	path = gtk_tree_model_get_path (model, &reldata->primary_iter);
	gtk_tree_model_row_changed (model, path, &reldata->primary_iter);
	gtk_tree_path_free (path);
	model = GTK_TREE_MODEL (reldata->secondary->store);
	path = gtk_tree_model_get_path (model, &reldata->secondary_iter);
	gtk_tree_model_row_changed (model, path, &reldata->secondary_iter);
	gtk_tree_path_free (path);
}

void	ug_category_gtk_clear (UgCategoryGtk* cgtk, UgCategoryHints hint, guint from_nth)
{
	UgDataset*		dataset;
	GtkTreeModel*	model;
	GtkTreeIter		iter;
	gboolean		valid;

	if (hint & UG_HINT_RECYCLED)
		model = cgtk->recycled.model;
	else if (hint & UG_HINT_FINISHED)
		model = cgtk->finished.model;
	else
		model = cgtk->queuing.model;

	valid = gtk_tree_model_iter_nth_child (model, &iter, NULL, from_nth);
	while (valid) {
		gtk_tree_model_get (model, &iter, 0, &dataset, -1);
		valid = gtk_tree_model_iter_next (model, &iter);
		ug_category_gtk_remove (cgtk, dataset);
	}
}

void	ug_category_gtk_move_to (UgCategoryGtk* cgtk, UgDataset* dataset, UgCategoryGtk* cgtk_dest)
{
	UgRelationGtk*		relation;

	relation = UG_DATASET_RELATION_GTK (dataset);
	if (relation->primary == cgtk_dest || relation->secondary == cgtk_dest)
		return;

	if (relation->secondary) {
		cgtk = relation->secondary;
		// stop active job in secondary category
		g_queue_remove (&cgtk->active, dataset);
		ug_category_stop ((UgCategory*) cgtk, dataset);
		// remove
		gtk_list_store_remove (cgtk->store, &relation->secondary_iter);
		// add
		relation->secondary = cgtk_dest;
		gtk_list_store_append (cgtk_dest->store, &relation->secondary_iter);
		gtk_list_store_set (cgtk_dest->store, &relation->secondary_iter, 0, dataset, -1);
		// switch
		ug_category_gtk_switch (cgtk_dest, dataset, relation->hints);
	}
}

GList*	ug_category_gtk_get_list (UgCategoryGtk* cgtk)
{
	UgDataset*		dataset;
	GtkTreeModel*	model;
	GtkTreeIter		iter;
	GList*			list;
	gboolean		valid;

	list = NULL;
	model = GTK_TREE_MODEL (cgtk->store);
	valid = gtk_tree_model_get_iter_first (model, &iter);
	while (valid) {
		gtk_tree_model_get (model, &iter, 0, &dataset, -1);
		list = g_list_prepend (list, dataset);
		valid = gtk_tree_model_iter_next (model, &iter);
	}
	return g_list_reverse (list);
}

gboolean	ug_category_gtk_move_selected_up (UgCategoryGtk* cgtk, UgDownloadWidget* dwidget)
{
	UgDataset*		dataset;
	UgRelationGtk*	relation;
	UgRelationGtk*	relation_prev;
	GtkTreeIter		iter;
	GList*			list;
	GList*			link;
	gint			index;
	gint			index_prev;
	gboolean		result = FALSE;

	index_prev = -1;
	list = ug_download_widget_get_selected_indices (dwidget);
	// scroll to first item
	if (list) {
		index = GPOINTER_TO_INT (list->data);
		if (index > 0)
			index--;
		gtk_tree_view_scroll_to_point (dwidget->view, -1, index);
	}
	// move
	for (link = list;  link;  link = link->next) {
		index = GPOINTER_TO_INT (link->data);
		if (index == index_prev+1) {
			index_prev++;
			continue;
		}
		index_prev = index - 1;
		// get previous
		gtk_tree_model_iter_nth_child (dwidget->model, &iter, NULL, index_prev);
		gtk_tree_model_get (dwidget->model, &iter, 0, &dataset, -1);
		relation_prev = UG_DATASET_RELATION_GTK (dataset);
		// get current
		gtk_tree_model_iter_next (dwidget->model, &iter);
		gtk_tree_model_get (dwidget->model, &iter, 0, &dataset, -1);
		relation = UG_DATASET_RELATION_GTK (dataset);
		// swap
		gtk_list_store_swap (relation->primary->store,
				&relation->primary_iter, &relation_prev->primary_iter);
		gtk_list_store_swap (relation->secondary->store,
				&relation->secondary_iter, &relation_prev->secondary_iter);
		result = TRUE;
	}

	g_list_free (list);
	return result;
}

gboolean	ug_category_gtk_move_selected_down (UgCategoryGtk* cgtk, UgDownloadWidget* dwidget)
{
	UgDataset*		dataset;
	UgRelationGtk*	relation;
	UgRelationGtk*	relation_next;
	GtkTreeIter		iter;
	GList*			list;
	GList*			link;
	gint			index;
	gint			index_next;
	gboolean		result = FALSE;

	index_next = gtk_tree_model_iter_n_children (dwidget->model, NULL);
	list = ug_download_widget_get_selected_indices (dwidget);
	link = g_list_last (list);
	// scroll to last item
	if (link) {
		index = GPOINTER_TO_INT (link->data);
		if (index < index_next -1)
			index++;
		gtk_tree_view_scroll_to_point (dwidget->view, -1, index);
	}
	// move
	for (;  link;  link = link->prev) {
		index = GPOINTER_TO_INT (link->data);
		if (index == index_next-1) {
			index_next--;
			continue;
		}
		index_next = index + 1;
		// get current
		gtk_tree_model_iter_nth_child (dwidget->model, &iter, NULL, index);
		gtk_tree_model_get (dwidget->model, &iter, 0, &dataset, -1);
		relation = UG_DATASET_RELATION_GTK (dataset);
		// get next
		gtk_tree_model_iter_next (dwidget->model, &iter);
		gtk_tree_model_get (dwidget->model, &iter, 0, &dataset, -1);
		relation_next = UG_DATASET_RELATION_GTK (dataset);
		// swap
		gtk_list_store_swap (relation->primary->store,
				&relation->primary_iter, &relation_next->primary_iter);
		gtk_list_store_swap (relation->secondary->store,
				&relation->secondary_iter, &relation_next->secondary_iter);
		result = TRUE;
	}

	g_list_free (list);
	return result;
}

gboolean	ug_category_gtk_move_selected_to_top (UgCategoryGtk* cgtk, UgDownloadWidget* dwidget)
{
	UgDataset*		dataset;
	UgRelationGtk*	relation;
	UgRelationGtk*	relation_top;
	GtkTreeIter		iter;
	GList*			list;
	GList*			link;
	gint			index;
	gint			index_top;

	// get movable jobs
	relation_top = NULL;
	index_top = 0;
	list = ug_download_widget_get_selected_indices (dwidget);
	for (link=list;  link;  link=link->next, index_top++) {
		index = GPOINTER_TO_INT (link->data);
		if (index == index_top) {
			link->data = NULL;
			continue;
		}
		else {
			gtk_tree_model_iter_nth_child (dwidget->model, &iter, NULL, index);
			gtk_tree_model_get (dwidget->model, &iter, 0, &dataset, -1);
			link->data = UG_DATASET_RELATION_GTK (dataset);
		}
		if (relation_top == NULL) {
			gtk_tree_model_iter_nth_child (dwidget->model, &iter, NULL, index_top);
			gtk_tree_model_get (dwidget->model, &iter, 0, &dataset, -1);
			relation_top = UG_DATASET_RELATION_GTK (dataset);
		}
	}
	list = g_list_remove_all (list, NULL);
	if (list == NULL)
		return FALSE;

	// move to top
	for (link = list;  link;  link = link->next) {
		relation = link->data;
		gtk_list_store_move_before (relation->primary->store,
				&relation->primary_iter, &relation_top->primary_iter);
		gtk_list_store_move_before (relation->secondary->store,
				&relation->secondary_iter, &relation_top->secondary_iter);
	}
	g_list_free (list);

	// scroll to top
	gtk_tree_view_scroll_to_point (dwidget->view, -1, 0);
	return TRUE;
}

gboolean	ug_category_gtk_move_selected_to_bottom (UgCategoryGtk* cgtk, UgDownloadWidget* dwidget)
{
	UgDataset*		dataset;
	UgRelationGtk*	relation;
	UgRelationGtk*	relation_bottom;
	GtkTreeIter		iter;
	GList*			list;
	GList*			link;
	gint			index;
	gint			index_bottom;

	// get movable jobs
	relation_bottom = NULL;
	index_bottom = gtk_tree_model_iter_n_children (dwidget->model, NULL) - 1;
	list = ug_download_widget_get_selected_indices (dwidget);
	list = g_list_reverse (list);
	for (link=list;  link;  link=link->next, index_bottom--) {
		index = GPOINTER_TO_INT (link->data);
		if (index == index_bottom) {
			link->data = NULL;
			continue;
		}
		else {
			gtk_tree_model_iter_nth_child (dwidget->model, &iter, NULL, index);
			gtk_tree_model_get (dwidget->model, &iter, 0, &dataset, -1);
			link->data = UG_DATASET_RELATION_GTK (dataset);
		}
		if (relation_bottom == NULL) {
			gtk_tree_model_iter_nth_child (dwidget->model, &iter, NULL, index_bottom);
			gtk_tree_model_get (dwidget->model, &iter, 0, &dataset, -1);
			relation_bottom = UG_DATASET_RELATION_GTK (dataset);
		}
	}
	list = g_list_remove_all (list, NULL);
	if (list == NULL)
		return FALSE;

	// move to bottom
	for (link = list;  link;  link = link->next) {
		relation = link->data;
		gtk_list_store_move_after (relation->primary->store,
				&relation->primary_iter, &relation_bottom->primary_iter);
		gtk_list_store_move_after (relation->secondary->store,
				&relation->secondary_iter, &relation_bottom->secondary_iter);
	}
	g_list_free (list);

	// scroll to bottom
	index = gtk_tree_model_iter_n_children (dwidget->model, NULL) -1;
	gtk_tree_view_scroll_to_point (dwidget->view, -1, index);
	return TRUE;
}

gboolean	ug_category_gtk_start_1 (UgCategoryGtk* cgtk, UgDataset* dataset)
{
	if (ug_category_start ((UgCategory*) cgtk, dataset)) {
		if (g_queue_find (&cgtk->active, dataset) == NULL)
			g_queue_push_tail (&cgtk->active, dataset);
		return TRUE;
	}
	return FALSE;
}

void	ug_category_gtk_stop_1 (UgCategoryGtk* cgtk, UgDataset* dataset)
{
	ug_category_stop ((UgCategory*) cgtk, dataset);
	g_queue_remove (&cgtk->active, dataset);
}

// returned flag
// 0     : status no changed.
// 1 << 0: status changed. (start, stop, move, remove)
// 1 << 1: status changed. some job stop by user
guint	ug_category_gtk_refresh (UgCategoryGtk* cgtk, GList** completed)
{
	UgDataset*		dataset;
	UgRelationGtk*	relation;
	GList*			link;
	guint			result = 0;

	for (link = cgtk->active.head;  link;  link = link->next) {
		dataset = link->data;
		// refresh status & data
		if (ug_category_refresh ((UgCategory*) cgtk, dataset))
			continue;
		// stop & clear
		ug_category_stop ((UgCategory*) cgtk, dataset);
		link->data = NULL;
		// relation
		relation = UG_DATASET_RELATION_GTK (dataset);
		result |= 1;
		if (relation->user_action) {
			relation->user_action = FALSE;	// clear status "stop by user"
			result |= (1 << 1);
		}
		if (relation->hints & UG_HINT_FINISHED) {
			ug_category_gtk_switch (cgtk, dataset, UG_HINT_FINISHED);
			if (completed)
				*completed = g_list_prepend (*completed, dataset);
		}
	}
	g_queue_remove_all (&cgtk->active, NULL);

	return result;
}

guint	ug_category_gtk_activate (UgCategoryGtk* cgtk)
{
	UgDataset*		dataset;
	UgRelationGtk*	relation;
	GtkTreeIter		iter;
	gboolean		valid;
	guint			result = 0;

	valid = gtk_tree_model_get_iter_first (cgtk->queuing.model, &iter);
	while (valid  &&  cgtk->active.length < cgtk->active_limit) {
		gtk_tree_model_get (cgtk->queuing.model, &iter, 0, &dataset, -1);
		valid = gtk_tree_model_iter_next (cgtk->queuing.model, &iter);
		// check & start
		relation = UG_DATASET_RELATION_GTK (dataset);
		if (relation->hints & (UG_HINT_UNRUNNABLE | UG_HINT_ACTIVE))
			continue;
		result |= 1;
		if (ug_category_start ((UgCategory*) cgtk, dataset) == FALSE) {
			g_free (relation->message);
			relation->message_type = UG_MESSAGE_ERROR;
			relation->message = g_strdup (_("Unsupported scheme (protocol)."));
			continue;
		}
		g_queue_push_tail (&cgtk->active, dataset);
	}
	return result;
}

guint	ug_category_gtk_clear_excess (UgCategoryGtk* cgtk)
{
	guint	n;
	guint	result = 0;

	n = gtk_tree_model_iter_n_children (cgtk->finished.model, NULL);
	if (n > cgtk->finished_limit)
		result |= 1;
	n = gtk_tree_model_iter_n_children (cgtk->recycled.model, NULL);
	if (n > cgtk->recycled_limit)
		result |= 1;

	ug_category_gtk_clear (cgtk, UG_HINT_FINISHED, cgtk->finished_limit);
	ug_category_gtk_clear (cgtk, UG_HINT_RECYCLED, cgtk->recycled_limit);

	return result;
}

// ----------------------------------------------
// static function
static gboolean ug_category_gtk_watch (UgCategoryGtk* ccmd, UgDataset* dataset, UgMessage* message, gpointer user_data)
{
	UgRelationGtk*	relation;

	relation = UG_DATASET_RELATION_GTK (dataset);

	switch (message->type) {
	case UG_MESSAGE_ERROR:
	case UG_MESSAGE_WARNING:
		g_free (relation->message);
		relation->message = g_strdup (message->string);
		relation->message_type = message->type;
		break;

	default:
		break;
	}
	return FALSE;
}

// Function used by GtkTreeModelSort.
static gint	ug_download_model_compare (GtkTreeModel* model, GtkTreeIter* a, GtkTreeIter* b, gpointer user_data)
{
	UgDataset*		dataset;
	UgDataCommon*	common;
	gchar*			name1;
	gchar*			name2;

	gtk_tree_model_get (model, a, 0, &dataset, -1);
	if (dataset == NULL)
		return 0;
	common = UG_DATASET_COMMON (dataset);
	if (common->name)
		name1 = common->name;
	else if (common->file)
		name1 = common->file;
	else
		name1 = "unnamed";

	gtk_tree_model_get (model, b, 0, &dataset, -1);
	if (dataset == NULL)
		return 0;
	common = UG_DATASET_COMMON (dataset);
	if (common->name)
		name2 = common->name;
	else if (common->file)
		name2 = common->file;
	else
		name2 = "unnamed";
	return strcmp (name1, name2);
}

// Function used by GtkTreeModelFilter.
static gboolean	ug_download_model_filter_queuing  (GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	UgDataset*		dataset;
	UgRelation*		relation;

	gtk_tree_model_get (model, iter, 0, &dataset, -1);
	if (dataset) {
		relation = UG_DATASET_RELATION (dataset);
		if (relation->hints & (UG_HINT_FINISHED | UG_HINT_RECYCLED))
			return FALSE;
	}
	return TRUE;
}

// Function used by GtkTreeModelFilter.
static gboolean	ug_download_model_filter_finished (GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	UgDataset*		dataset;
	UgRelation*		relation;

	gtk_tree_model_get (model, iter, 0, &dataset, -1);
	if (dataset) {
		relation = UG_DATASET_RELATION (dataset);
		if (relation->hints & UG_HINT_FINISHED)
			return TRUE;
	}
	return FALSE;
}

// Function used by GtkTreeModelFilter.
static gboolean	ug_download_model_filter_recycled (GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	UgDataset*		dataset;
	UgRelation*		relation;

	gtk_tree_model_get (model, iter, 0, &dataset, -1);
	if (dataset) {
		relation = UG_DATASET_RELATION (dataset);
		if (relation->hints & UG_HINT_RECYCLED)
			return TRUE;
	}
	return FALSE;
}

// Function used by ug_category_gtk_switch().
static UgDataset* ug_download_model_get_last (GtkTreeModel* model)
{
	GtkTreeIter	iter;
	UgDataset*	dataset;
	guint		length;

	length = gtk_tree_model_iter_n_children (model, NULL);
	if (length > 0) {
		gtk_tree_model_iter_nth_child (model, &iter, NULL, length - 1);
		gtk_tree_model_get (model, &iter, 0, &dataset, -1);
		return dataset;
	}
	return NULL;
}

// Function used by ug_category_gtk_switch().
static UgDataset* ug_download_model_get_first (GtkTreeModel* model)
{
	GtkTreeIter	iter;
	UgDataset*	dataset;

	if (gtk_tree_model_get_iter_first (model, &iter)) {
		gtk_tree_model_get (model, &iter, 0, &dataset, -1);
		return dataset;
	}
	return NULL;
}


// ----------------------------------------------------------------------------
// UgRelationGtk
//
static void	ug_relation_gtk_assign (UgRelation* relation, UgRelation* src);

static UgDataEntry	relation_gtk_data_entry[] =
{
	UG_RELATION_DATA_ENTRY,
	{NULL},			// null-terminated
};

static UgDataClass	relation_gtk_data_class =
{
	UG_RELATION_CLASS_NAME,		// name
	NULL,						// reserve
	sizeof (UgRelationGtk),		// instance_size
	relation_gtk_data_entry,	// entry

	(UgInitFunc)     NULL,
	(UgFinalizeFunc) NULL,
	(UgAssignFunc)   ug_relation_gtk_assign,
};
// extern
const	UgDataClass*	UgRelationGtkClass = &relation_gtk_data_class;

static void	ug_relation_gtk_assign (UgRelation* relation, UgRelation* src)
{
	if (src->hints & UG_HINT_PAUSED) {
		relation->hints |=  UG_HINT_PAUSED;
		relation->hints &= ~UG_HINT_ACTIVE;
	}
	else
		relation->hints &= ~UG_HINT_PAUSED;
}

